// external lib
import { ListenerRss, ListenerRSSInfos } from "listener-rss";

// local lib
import { ManageListener } from "../";

// Unit test
import path from "path";
import events from "events";
import nock from "nock";
import * as chai from "chai";
import sinon from "ts-sinon";
import sinonChai from "sinon-chai";

chai.use(sinonChai);

const expect = chai.expect;

describe("test class GestionListener", function () {
  describe("test constructor", function () {
    it("should create a new manageListener object without params", function () {
      // when
      const ml = new ManageListener();

      // expect
      expect(ml.listenerArray).to.be.empty;
    });

    it("should create a new manageListener object with a param file", function () {
      // given
      const tmp = require("./RessourcesTest/FakeRessources/save.json");

      // when
      const ml = new ManageListener({
        path: "tests/RessourcesTest/FakeRessources/save.json",
      });

      // expect
      expect(tmp).to.be.eql(ml.listenerArray.map((item) => item.getProperty()));
    });
  });

  describe("test saving", function () {
    beforeEach("Clean old files if exist", function () {
      const fs = require("fs");
      if (fs.existsSync("tests/RessourcesTest/FakeRessources/fake_save.json"))
        fs.unlinkSync("tests/RessourcesTest/FakeRessources/fake_save.json");
    });

    afterEach("Clean old files if exist", function () {
      const fs = require("fs");
      if (fs.existsSync("tests/RessourcesTest/FakeRessources/fake_save.json"))
        fs.unlinkSync("tests/RessourcesTest/FakeRessources/fake_save.json");
    });

    it("should save the inserted objects", function (): void {
      // given
      const ml = new ManageListener();
      let expectedObj: ListenerRss[] = [];

      for (let i = 1; i < 10; i++) {
        const generatedInfo = {
          address: `fake.rss.service.${i}`,
          customfields: {
            "media:group": ["media:group"],
          },
          timeloop: i * 10,
        };
        expectedObj.push(new ListenerRss(generatedInfo));
        ml.addNewListener(generatedInfo);
      }

      // when
      ml.save("tests/RessourcesTest/FakeRessources/fake_save.json");
      const tmp = require("./RessourcesTest/FakeRessources/save.json");

      ml.save("tests/RessourcesTest/FakeRessources/fake_save.json");

      ml.save("tests/RessourcesTest/FakeRessources/fake_save.json");

      // expect
      expect(expectedObj.map((item) => item.getProperty())).to.be.eql(tmp);
    });
  });

  describe("test events", function () {
    beforeEach(function () {
      nock.disableNetConnect();
    });

    afterEach(function () {
      nock.cleanAll();
    });

    it("should call correctly the events (no aggregation)", async function () {
      // given
      const clock = sinon.useFakeTimers();
      const ml = new ManageListener({
        path: "tests/RessourcesTest/RealRessources/save_no_history.json",
      });

      const updateSpy = sinon.spy();
      const newEntriesSpy = sinon.spy();

      const tabChannelId = [
        "UCOuIgj0CYCXCvjWywjDbauw",
        "UCh2YBKhYIy-_LtfCIn2Jycg",
        "UCY7klexd1qEqxgqYK6W7BVQ",
      ];

      ml.on("update", updateSpy);
      ml.on("newEntries", newEntriesSpy);
      tabChannelId.forEach((item: string) => {
        nock("https://www.youtube.com")
          .get(`/feeds/videos.xml?channel_id=${item}`)
          .replyWithFile(
            200,
            path.join(__dirname, `RessourcesTest/RealRessources/${item}.rss`),
            { "content-type": "text/xml", charset: "utf-8" }
          )
          .persist();
      });
      ml.startAll();

      // here we're testing the first call
      // when
      await Promise.all([
        events.once(ml.listenerArray[0], "update"),
        events.once(ml.listenerArray[1], "update"),
        events.once(ml.listenerArray[2], "update"),
      ]);

      //expect
      expect(updateSpy).to.have.been.calledThrice;
      expect(newEntriesSpy).to.have.been.calledThrice;

      // here we're testing the second call for the first listener
      // given
      updateSpy.resetHistory();
      newEntriesSpy.resetHistory();

      // when
      await clock.tickAsync(10000);
      await events.once(ml.listenerArray[0], "update");

      //expect
      expect(updateSpy).to.have.been.calledOnce;
      expect(newEntriesSpy).to.not.have.been.called;

      // here we're testing the thrid call for the first listener and second for the second listener
      // given
      updateSpy.resetHistory();

      // when
      await clock.tickAsync(10000);
      await Promise.all([
        events.once(ml.listenerArray[0], "update"),
        events.once(ml.listenerArray[1], "update"),
      ]);

      //expect
      expect(updateSpy).to.have.been.calledTwice;
      expect(newEntriesSpy).to.not.have.been.called;

      // here we're testing the fourth call for the first listener and second for the third listener
      // given
      updateSpy.resetHistory();

      // when
      await clock.tickAsync(10000);
      await Promise.all([
        events.once(ml.listenerArray[0], "update"),
        events.once(ml.listenerArray[2], "update"),
      ]);

      //expect
      expect(updateSpy).to.have.been.calledTwice;
      expect(newEntriesSpy).to.not.have.been.called;

      ml.stopAll();
    });

    it("should call correctly the events with a new entry (no aggregation)", async function () {
      // given
      const clock = sinon.useFakeTimers();
      const ml = new ManageListener({
        path: "tests/RessourcesTest/RealRessources/save.json",
      });

      const updateSpy = sinon.spy();
      const newEntriesSpy = sinon.spy();

      const tabChannelId = [
        "UCOuIgj0CYCXCvjWywjDbauw",
        "UCh2YBKhYIy-_LtfCIn2Jycg",
        "UCY7klexd1qEqxgqYK6W7BVQ",
      ];

      ml.on("update", updateSpy);
      ml.on("newEntries", newEntriesSpy);
      tabChannelId.forEach((item: string) => {
        nock("https://www.youtube.com")
          .get(`/feeds/videos.xml?channel_id=${item}`)
          .once()
          .replyWithFile(
            200,
            path.join(__dirname, `RessourcesTest/RealRessources/${item}.rss`),
            { "content-type": "text/xml", charset: "utf-8" }
          );
      });
      ml.startAll();

      // here we're testing the first call
      // when
      await Promise.all([
        events.once(ml.listenerArray[0], "update"),
        events.once(ml.listenerArray[1], "update"),
        events.once(ml.listenerArray[2], "update"),
      ]);

      //expect
      expect(updateSpy).to.have.been.calledThrice;
      expect(newEntriesSpy).to.not.have.been.called;

      // here we're testing the second call for the first listener
      // given
      updateSpy.resetHistory();
      newEntriesSpy.resetHistory();

      tabChannelId.forEach((item: string) => {
        nock("https://www.youtube.com")
          .get(`/feeds/videos.xml?channel_id=${item}`)
          .replyWithFile(
            200,
            path.join(
              __dirname,
              `RessourcesTest/RealRessources/WithUpdate/${item}.rss`
            ),
            { "content-type": "text/xml", charset: "utf-8" }
          )
          .persist();
      });

      // when
      await clock.tickAsync(10000);
      await events.once(ml.listenerArray[0], "update");

      //expect
      expect(updateSpy).to.have.been.calledOnce;
      expect(newEntriesSpy).to.have.been.calledOnce;

      // here we're testing the thrid call for the first listener and second for the second listener
      // given
      updateSpy.resetHistory();
      newEntriesSpy.resetHistory();

      // when
      await clock.tickAsync(10000);
      await Promise.all([
        events.once(ml.listenerArray[0], "update"),
        events.once(ml.listenerArray[1], "update"),
      ]);

      //expect
      expect(updateSpy).to.have.been.calledTwice;
      expect(newEntriesSpy).to.have.been.calledOnce;

      // here we're testing the fourth call for the first listener and second for the third listener
      // given
      updateSpy.resetHistory();
      newEntriesSpy.resetHistory();

      // when
      await clock.tickAsync(10000);
      await Promise.all([
        events.once(ml.listenerArray[0], "update"),
        events.once(ml.listenerArray[2], "update"),
      ]);

      //expect
      expect(updateSpy).to.have.been.calledTwice;
      expect(newEntriesSpy).to.have.been.calledOnce;

      ml.stopAll();
    });

    it("should call correctly the events (with aggregation)", async function () {
      // given
      const clock = sinon.useFakeTimers();
      const ml = new ManageListener({
        timeloop: 30,
        path: "tests/RessourcesTest/RealRessources/save_no_history.json",
      });

      const updateSpy = sinon.spy((obj) => console.log(obj));
      const newEntriesSpy = sinon.spy();

      const tabChannelId = [
        "UCOuIgj0CYCXCvjWywjDbauw",
        "UCh2YBKhYIy-_LtfCIn2Jycg",
        "UCY7klexd1qEqxgqYK6W7BVQ",
      ];

      ml.on("update", updateSpy);
      ml.on("newEntries", newEntriesSpy);
      tabChannelId.forEach((item: string) => {
        nock("https://www.youtube.com")
          .get(`/feeds/videos.xml?channel_id=${item}`)
          .replyWithFile(
            200,
            path.join(__dirname, `RessourcesTest/RealRessources/${item}.rss`),
            { "content-type": "text/xml", charset: "utf-8" }
          )
          .persist();
      });
      ml.startAll();

      // when
      await Promise.all([
        events.once(ml.listenerArray[0], "update"),
        events.once(ml.listenerArray[1], "update"),
        events.once(ml.listenerArray[2], "update"),
      ]);

      //expect
      expect(updateSpy).to.have.been.calledOnce;
      expect(newEntriesSpy).to.have.been.calledOnce;

      // given
      updateSpy.resetHistory();
      newEntriesSpy.resetHistory();

      // when
      await clock.tickAsync(10000);
      await events.once(ml.listenerArray[0], "update");

      //expect
      expect(updateSpy).to.not.have.been.called;
      expect(newEntriesSpy).to.not.have.been.called;

      // when
      await clock.tickAsync(10000);
      await Promise.all([
        events.once(ml.listenerArray[0], "update"),
        events.once(ml.listenerArray[1], "update"),
      ]);

      //expect
      expect(updateSpy).to.not.have.been.called;
      expect(newEntriesSpy).to.not.have.been.called;

      // when
      await clock.tickAsync(10000);
      await Promise.all([
        events.once(ml.listenerArray[0], "update"),
        events.once(ml.listenerArray[2], "update"),
      ]);

      //expect
      expect(updateSpy).to.have.been.calledOnce;
      expect(newEntriesSpy).to.not.have.been.called;

      ml.stopAll();
    });

    it("should call correctly the events with a new entry (with aggregation)", async function () {
      // given
      const clock = sinon.useFakeTimers();
      const ml = new ManageListener({
        timeloop: 30,
        path: "tests/RessourcesTest/RealRessources/save.json",
      });

      const updateSpy = sinon.spy();
      const newEntriesSpy = sinon.spy();

      const tabChannelId = [
        "UCOuIgj0CYCXCvjWywjDbauw",
        "UCh2YBKhYIy-_LtfCIn2Jycg",
        "UCY7klexd1qEqxgqYK6W7BVQ",
      ];

      ml.on("update", updateSpy);
      ml.on("newEntries", newEntriesSpy);
      tabChannelId.forEach((item: string) => {
        nock("https://www.youtube.com")
          .get(`/feeds/videos.xml?channel_id=${item}`)
          .once()
          .replyWithFile(
            200,
            path.join(__dirname, `RessourcesTest/RealRessources/${item}.rss`),
            { "content-type": "text/xml", charset: "utf-8" }
          );
      });
      ml.startAll();

      // when
      await Promise.all([
        events.once(ml.listenerArray[0], "update"),
        events.once(ml.listenerArray[1], "update"),
        events.once(ml.listenerArray[2], "update"),
      ]);

      //expect
      expect(updateSpy).to.have.been.calledOnce;
      expect(newEntriesSpy).to.not.have.been.called;

      // given
      updateSpy.resetHistory();

      tabChannelId.forEach((item: string) => {
        nock("https://www.youtube.com")
          .get(`/feeds/videos.xml?channel_id=${item}`)
          .replyWithFile(
            200,
            path.join(
              __dirname,
              `RessourcesTest/RealRessources/WithUpdate/${item}.rss`
            ),
            { "content-type": "text/xml", charset: "utf-8" }
          )
          .persist();
      });

      // when
      await clock.tickAsync(10000);
      await events.once(ml.listenerArray[0], "update");

      //expect
      expect(updateSpy).to.not.have.been.called;
      expect(newEntriesSpy).to.not.have.been.called;

      // when
      await clock.tickAsync(10000);
      await Promise.all([
        events.once(ml.listenerArray[0], "update"),
        events.once(ml.listenerArray[1], "update"),
      ]);

      //expect
      expect(updateSpy).to.not.have.been.called;
      expect(newEntriesSpy).to.not.have.been.called;

      // when
      await clock.tickAsync(10000);
      await Promise.all([
        events.once(ml.listenerArray[0], "update"),
        events.once(ml.listenerArray[2], "update"),
      ]);

      //expect
      expect(updateSpy).to.have.been.calledOnce;
      expect(newEntriesSpy).to.have.been.calledOnce;

      ml.stopAll();
    });
  });
});
